-------------------------------------------------------------------------------
--! @file       GF2_32_package.vhd
--! @brief      GF(2^32) functions for TriviA0v2
--! @project    CAESAR Candidate Evaluation
--! @author     Michael X. Lyons
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------


LIBRARY ieee;
USE ieee.std_logic_1164.ALL;


PACKAGE GF2_32_package is
    -- multiply a 32-bit value by 'x' in the finite field
    FUNCTION gf2_32_by_x(a : STD_LOGIC_VECTOR(31 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR;

    -- multiply a 32-bit value by 'x^2' in the finite field
    FUNCTION gf2_32_by_x2(a : STD_LOGIC_VECTOR(31 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR;

    -- multiply a 32-bit value by 'x^4' in the finite field
    FUNCTION gf2_32_by_x4(a : STD_LOGIC_VECTOR(31 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR;

	-- primitive polynomial x^32 + x^22 + x^2 + x + 1 mod 2^32
    -- CONSTANT ALPHA_32 : UNSIGNED(31 DOWNTO 0) := (22 => '1',
    --                                                2 => '1',
    --                                                1 => '1',
    --                                                0 => '1',
    --                                               OTHERS => '0');
END PACKAGE;

PACKAGE BODY GF2_32_package is
    FUNCTION gf2_32_by_x(a : STD_LOGIC_VECTOR(31 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR IS

        variable r : std_logic_vector(31 downto 0);
    BEGIN
        -- shift 'a' 1 bit to the left with '0' fill on the right
        -- if the MSB prior to the shift was '1' XOR the shifted value with the primitive polynomial

        -- simple code
        -- IF a(31) = '0' THEN
        --     RETURN a(30 DOWNTO 0) & '0';
        -- ELSE
        --     RETURN (a(30 DOWNTO 0) & '0') XOR ALPHA_32;
        -- END IF;

        -- inline code
        -- RETURN (a(30 DOWNTO 0) & '0') xor
        --        ("000000000" & a(31) & "0000000000000000000" & a(31) & a(31) & a(31));

        -- save 29 of 32 XORs
        --RETURN  a(30 DOWNTO 22)  &
        --       (a(21) xor a(31)) &
        --        a(20 DOWNTO  2)  &
        --       (a( 1) xor a(31)) &
        --       (a( 0) xor a(31)) &
        --                  a(31);

        r(31 downto 23) := a(30 downto 22);
        r(22)           := a(21)           xor a(31);
        r(21 downto  3) := a(20 downto  2);
        r( 2)           := a( 1)           xor a(31);
        r( 1)           := a( 0)           xor a(31);
        r( 0)           :=                     a(31);
        return r;
    END FUNCTION;

    FUNCTION gf2_32_by_x2(a : STD_LOGIC_VECTOR(31 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR IS

        variable r : std_logic_vector(31 downto 0);
    BEGIN
        -- see above

        -- simple code
        -- RETURN gf2_32_by_x(gf2_32_by_x(a));

        -- inline code
        -- RETURN (a(29 DOWNTO 0) & "00") xor
        --        ("00000000" & a(31) & "0000000000000000000"  & a(31) & a(31) & a(31) &   '0') xor
        --        ("000000000" & a(30) & "0000000000000000000"         & a(30) & a(30) & a(30));

        -- save 58 of 64 XORs
        --RETURN  a(29 DOWNTO 22)            &
        --       (a(21) xor a(31)          ) &
        --       (a(20)           xor a(30)) &
        --        a(19 DOWNTO  2)            &
        --       (a( 1) xor a(31)          ) &
        --       (a( 0) xor a(31) xor a(30)) &
        --       (          a(31) xor a(30)) &
        --                            a(30);

        r(31 downto 24) := a(29 downto 22);
        r(23)           := a(21)           xor a(31);
        r(22)           := a(20)                     xor a(30);
        r(21 downto  4) := a(19 downto  2);
        r( 3)           := a( 1)           xor a(31);
        r( 2)           := a( 0)           xor a(31) xor a(30);
        r( 1)           :=                     a(31) xor a(30);
        r( 0)           :=                               a(30);
        return r;
    END FUNCTION;

    FUNCTION gf2_32_by_x4(a : STD_LOGIC_VECTOR(31 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR IS

        variable r : std_logic_vector(31 downto 0);
    BEGIN
        -- see above

        -- simple code
        -- RETURN gf2_32_by_x2(gf2_32_by_x2(a));

        -- inline code
        -- RETURN (a(28 DOWNTO 0) & "000") xor
        --        ("0000000" & a(31) & "0000000000000000000" & a(31) & a(31) & a(31) & "00"         ) xor
        --        ("00000000" & a(30) & "0000000000000000000"        & a(30) & a(30) & a(30) &   '0') xor
        --        ("000000000" & a(29) & "0000000000000000000"               & a(29) & a(29) & a(29));

        -- save 87 of 96 XORs
        --RETURN  a(28 DOWNTO 22)                      &
        --       (a(21) xor a(31)                    ) &
        --       (a(20)           xor a(30)          ) &
        --       (a(19)                     xor a(29)) &
        --        a(18 DOWNTO  2)                      &
        --       (a( 1) xor a(31)                    ) &
        --       (a( 0) xor a(31) xor a(30)          ) &
        --       (          a(31) xor a(30) xor a(29)) &
        --       (                    a(30) xor a(29)) &
        --                                      a(29);

        r(31 downto 25) := a(28 downto 22);
        r(24)           := a(21)           xor a(31);
        r(23)           := a(20)                     xor a(30);
        r(22)           := a(19)                               xor a(29);
        r(21 downto  5) := a(18 downto  2);
        r( 4)           := a( 1)           xor a(31);
        r( 3)           := a( 0)           xor a(31) xor a(30);
        r( 2)           :=                     a(31) xor a(30) xor a(29);
        r( 1)           :=                               a(30) xor a(29);
        r( 0)           :=                                         a(29);
        return r;
    END FUNCTION;
END PACKAGE BODY;