-------------------------------------------------------------------------------
--! @file       GF2_64_package.vhd
--! @brief      GF(2^64) functions for TriviA0v2
--! @project    CAESAR Candidate Evaluation
--! @author     Michael X. Lyons
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------


LIBRARY ieee;
USE ieee.std_logic_1164.ALL;


PACKAGE GF2_64_package is
    -- multiply a 64-bit value by 'x' in the finite field
    FUNCTION gf2_64_by_x(a : STD_LOGIC_VECTOR(63 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR;

    -- multiply a 64-bit value by 'x^2' in the finite field
    FUNCTION gf2_64_by_x2(a : STD_LOGIC_VECTOR(63 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR;

	-- primitive polynomial x^64 + x^4 + x^3 + x + 1 mod 2^64
    -- CONSTANT ALPHA_64 : UNSIGNED(63 DOWNTO 0) := (4 => '1',
    --                                               3 => '1',
    --                                               1 => '1',
    --                                               0 => '1',
    --                                               OTHERS => '0');
END PACKAGE;

PACKAGE BODY GF2_64_package is
    FUNCTION gf2_64_by_x(a : STD_LOGIC_VECTOR(63 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR IS

        variable r : std_logic_vector(63 downto 0);
    BEGIN
        -- shift 'a' 1 bit to the left with '0' fill on the right
        -- if the MSB prior to the shift was '1' XOR the shifted value with the primitive polynomial

        -- simple code
        -- IF a(63) = '0' THEN
        --     RETURN a(62 DOWNTO 0) & '0';
        -- ELSE
        --     RETURN (a(62 DOWNTO 0) & '0') XOR ALPHA_64;
        -- END IF;

        -- inline code
        -- RETURN (a(62 DOWNTO 0) & '0') xor
        --        ("00000000000000000000000000000000000000000000000000000000000" & a(63) & a(63) & '0' & a(63) & a(63));

        -- save 61 of 64 XORs
        --RETURN  a(62 DOWNTO 4)   &
        --       (a( 3) xor a(63)) &
        --       (a( 2) xor a(63)) &
        --        a( 1)            &
        --       (a( 0) xor a(63)) &
        --                  a(63);

        r(63 downto  5) := a(62 downto  4);
        r( 4)           := a( 3)           xor a(63);
        r( 3)           := a( 2)           xor a(63);
        r( 2)           := a( 1);
        r( 1)           := a( 0)           xor a(63);
        r( 0)           :=                     a(63);
        return r;
    END FUNCTION;

    FUNCTION gf2_64_by_x2(a : STD_LOGIC_VECTOR(63 DOWNTO 0))
        RETURN STD_LOGIC_VECTOR IS

        variable r : std_logic_vector(63 downto 0);
    BEGIN
        -- see above

        -- simple code
        -- RETURN gf2_64_by_x(gf2_64_by_x(a));

        -- inline code
        -- RETURN (a(61 DOWNTO 0) & "00") xor
        --        ("0000000000000000000000000000000000000000000000000000000000" & a(63) & a(63) &   '0' & a(63) & a(63) &   '0') xor
        --        ("00000000000000000000000000000000000000000000000000000000000"        & a(62) & a(62) &   '0' & a(62) & a(62));

        -- save 122 of 128 XORs
        --RETURN  a(61 DOWNTO 4)             &
        --       (a( 3) xor a(63))           &
        --       (a( 2) xor a(63) xor a(62)) &
        --       (a( 1)           xor a(62)) &
        --       (a( 0) xor a(63)          ) &
        --       (          a(63) xor a(62)) &
        --                            a(62);               

        r(63 downto  6) := a(61 downto  4);
        r( 5)           := a( 3)           xor a(63);
        r( 4)           := a( 2)           xor a(63) xor a(62);
        r( 3)           := a( 1)                     xor a(62);
        r( 2)           := a( 0)           xor a(63);
        r( 1)           :=                     a(63) xor a(62);
        r( 0)           :=                               a(62);
        return r;
    END FUNCTION;
END PACKAGE BODY;