-------------------------------------------------------------------------------
--! @file       Mult_in_GF2_32.vhd
--! @brief      GF(2^32) multiplication for TriviA0v2
--! @project    CAESAR Candidate Evaluation
--! @author     Michael X. Lyons
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
--! Mastrovito architecture with custom "irregular f-network"
-------------------------------------------------------------------------------


LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE work.GF2_32_package.ALL;


ENTITY Mult_in_GF2_32 IS
    PORT (
         a  : IN  STD_LOGIC_VECTOR(31 DOWNTO 0);
         b  : IN  STD_LOGIC_VECTOR(31 DOWNTO 0);
         ab : OUT STD_LOGIC_VECTOR(31 DOWNTO 0)
	     );
END ENTITY Mult_in_GF2_32;

-- This architecture uses a 33x32-bit array and a 32x32-bit array.
-- It's analagous to a conventional shift-and-add multiplier
--   but an XOR is required if the MSB shifted out is '1'.
--   (Mastrovito calls this "shift-and-add-on-overflow".)
ARCHITECTURE interleaved OF Mult_in_GF2_32 IS
    CONSTANT ALPHA_32 : STD_LOGIC_VECTOR(31 DOWNTO 0) := (22 => '1',
                                                           2 => '1',
                                                           1 => '1',
                                                           0 => '1',
                                                           OTHERS => '0');
	-- primitive polynomial x^32 + x^22 + x^2 + x + 1

    -- this array accumulates the shift-and-perhaps-xor results
    TYPE result_array IS ARRAY (0 TO 32) OF STD_LOGIC_VECTOR(31 DOWNTO 0);
    SIGNAL results : result_array;

    -- this array holds 'a' multiplied in the finite field by successive powers 1,2,4,8,16,... of the primitive element 'x'
    TYPE a_power_array IS ARRAY (0 to 31) OF STD_LOGIC_VECTOR(31 DOWNTO 0);
    SIGNAL a_powers : a_power_array;
BEGIN
    results(0) <= (OTHERS => '0');
	 
	a_powers(0) <= a;

    mult: FOR i IN 1 TO 32 GENERATE
        -- if the current bit of interest in 'b' is 1 then XOR the current power of 'x' into the result
        results(i) <= results(i - 1) WHEN b(i - 1) = '0' ELSE
	                  results(i - 1) XOR a_powers(i - 1);

        not_last: IF i < 32 GENERATE
            -- each doubled power is the prior power shifted, and XORed with the primitive element if the MSB was 1
            --a_powers(i) <= (a_powers(i - 1)(30 DOWNTO 0) & '0') WHEN a_powers(i - 1)(31) = '0' ELSE
	        --               (a_powers(i - 1)(30 DOWNTO 0) & '0') XOR ALPHA_32;
            a_powers(i) <= gf2_32_by_x(a_powers(i - 1));
        END GENERATE not_last;
        
    END GENERATE mult;

	ab <= results(32);
END ARCHITECTURE interleaved;




--------------------------------------------------------------------
LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE work.GF2_32_package.ALL;

-- See http://www.commsys.isy.liu.se/publications/Theses/Thesis:Mastrovito:PhD:1991-02-08.pdf Chapter 4
-- Much of the following code was originally generated by a custom C program
--   Sub-expressions and groupings of them were coded manually
architecture mxl_mastrovito of Mult_in_GF2_32 is
    alias a0 is a(0);
    alias a1 is a(1);
    alias a2 is a(2);
    alias a3 is a(3);
    alias a4 is a(4);
    alias a5 is a(5);
    alias a6 is a(6);
    alias a7 is a(7);
    alias a8 is a(8);
    alias a9 is a(9);
    alias a10 is a(10);
    alias a11 is a(11);
    alias a12 is a(12);
    alias a13 is a(13);
    alias a14 is a(14);
    alias a15 is a(15);
    alias a16 is a(16);
    alias a17 is a(17);
    alias a18 is a(18);
    alias a19 is a(19);
    alias a20 is a(20);
    alias a21 is a(21);
    alias a22 is a(22);
    alias a23 is a(23);
    alias a24 is a(24);
    alias a25 is a(25);
    alias a26 is a(26);
    alias a27 is a(27);
    alias a28 is a(28);
    alias a29 is a(29);
    alias a30 is a(30);
    alias a31 is a(31);

    alias b0 is b(0);
    alias b1 is b(1);
    alias b2 is b(2);
    alias b3 is b(3);
    alias b4 is b(4);
    alias b5 is b(5);
    alias b6 is b(6);
    alias b7 is b(7);
    alias b8 is b(8);
    alias b9 is b(9);
    alias b10 is b(10);
    alias b11 is b(11);
    alias b12 is b(12);
    alias b13 is b(13);
    alias b14 is b(14);
    alias b15 is b(15);
    alias b16 is b(16);
    alias b17 is b(17);
    alias b18 is b(18);
    alias b19 is b(19);
    alias b20 is b(20);
    alias b21 is b(21);
    alias b22 is b(22);
    alias b23 is b(23);
    alias b24 is b(24);
    alias b25 is b(25);
    alias b26 is b(26);
    alias b27 is b(27);
    alias b28 is b(28);
    alias b29 is b(29);
    alias b30 is b(30);
    alias b31 is b(31);

    alias ab0 is ab(0);
    alias ab1 is ab(1);
    alias ab2 is ab(2);
    alias ab3 is ab(3);
    alias ab4 is ab(4);
    alias ab5 is ab(5);
    alias ab6 is ab(6);
    alias ab7 is ab(7);
    alias ab8 is ab(8);
    alias ab9 is ab(9);
    alias ab10 is ab(10);
    alias ab11 is ab(11);
    alias ab12 is ab(12);
    alias ab13 is ab(13);
    alias ab14 is ab(14);
    alias ab15 is ab(15);
    alias ab16 is ab(16);
    alias ab17 is ab(17);
    alias ab18 is ab(18);
    alias ab19 is ab(19);
    alias ab20 is ab(20);
    alias ab21 is ab(21);
    alias ab22 is ab(22);
    alias ab23 is ab(23);
    alias ab24 is ab(24);
    alias ab25 is ab(25);
    alias ab26 is ab(26);
    alias ab27 is ab(27);
    alias ab28 is ab(28);
    alias ab29 is ab(29);
    alias ab30 is ab(30);
    alias ab31 is ab(31);

    signal a31_a21 : std_logic;
    signal a30_a20 : std_logic;
    signal a29_a19 : std_logic;
    signal a28_a18 : std_logic;
    signal a27_a17 : std_logic;
    signal a26_a16 : std_logic;
    signal a25_a15 : std_logic;
    signal a24_a14 : std_logic;
    signal a23_a13 : std_logic;
    signal a22_a12 : std_logic;
    signal a21_a11 : std_logic;
    --signal a20_a10 : std_logic;  -- never used
    signal a19_a9 : std_logic;
    signal a18_a8 : std_logic;
    signal a17_a7 : std_logic;
    signal a16_a6 : std_logic;
    signal a15_a5 : std_logic;
    signal a14_a4 : std_logic;
    signal a13_a3 : std_logic;
    signal a12_a2 : std_logic;
    signal a11_a1 : std_logic;
    signal a10_a0 : std_logic;
    signal a31_a21_a11 : std_logic;
    signal a30_a20_a10 : std_logic;
    signal a29_a19_a9 : std_logic;
    signal a28_a18_a8 : std_logic;
    signal a27_a17_a7 : std_logic;
    signal a26_a16_a6 : std_logic;
    signal a25_a15_a5 : std_logic;
    signal a24_a14_a4 : std_logic;
    signal a23_a13_a3 : std_logic;
    signal a22_a12_a2 : std_logic;
    signal a21_a11_a1 : std_logic;
    signal a31_a0 : std_logic;
    signal a31_a30 : std_logic;
    signal a30_a29 : std_logic;
    signal a29_a28 : std_logic;
    signal a28_a27 : std_logic;
    signal a27_a26 : std_logic;
    signal a26_a25 : std_logic;
    signal a25_a24 : std_logic;
    signal a24_a23 : std_logic;
    signal a23_a22 : std_logic;
    signal a22_a21 : std_logic;
    signal a21_a20 : std_logic;
    signal a20_a19 : std_logic;
    signal a19_a18 : std_logic;
    signal a18_a17 : std_logic;
    signal a17_a16 : std_logic;
    signal a16_a15 : std_logic;
    signal a15_a14 : std_logic;
    signal a14_a13 : std_logic;
    signal a13_a12 : std_logic;
    signal a12_a11 : std_logic;
    signal a11_a10 : std_logic;
    signal a10_a9 : std_logic;
    signal a9_a8 : std_logic;
    signal a8_a7 : std_logic;
    signal a7_a6 : std_logic;
    signal a6_a5 : std_logic;
    signal a5_a4 : std_logic;
    signal a4_a3 : std_logic;
    signal a3_a2 : std_logic;
    signal a2_a1 : std_logic;
    --signal a1_a0 : std_logic;  -- never used
    signal a31_a22_a21 : std_logic;
    signal a31_a30_a21_a20 : std_logic;
    signal a30_a29_a20_a19 : std_logic;
    signal a29_a28_a19_a18 : std_logic;
    signal a28_a27_a18_a17 : std_logic;
    signal a27_a26_a17_a16 : std_logic;
    signal a26_a25_a16_a15 : std_logic;
    signal a25_a24_a15_a14 : std_logic;
    signal a24_a23_a14_a13 : std_logic;
    signal a23_a22_a13_a12 : std_logic;
    signal a31_a22_a21_a12_a11 : std_logic;
    signal a31_a30_a21_a20_a11_a10 : std_logic;
    signal a30_a29_a20_a19_a10_a9 : std_logic;
    signal a29_a28_a19_a18_a9_a8 : std_logic;
    signal a28_a27_a18_a17_a8_a7 : std_logic;
    signal a27_a26_a17_a16_a7_a6 : std_logic;
    signal a26_a25_a16_a15_a6_a5 : std_logic;
    signal a25_a24_a15_a14_a5_a4 : std_logic;
    signal a24_a23_a14_a13_a4_a3 : std_logic;
    signal a23_a22_a13_a12_a3_a2 : std_logic;
    signal a22_a21_a12_a11_a2_a1 : std_logic;
    signal a31_a1 : std_logic;
    signal a31_a30_a0 : std_logic;
    signal a31_a30_a29 : std_logic;
    signal a30_a29_a28 : std_logic;
    signal a29_a28_a27 : std_logic;
    signal a28_a27_a26 : std_logic;
    signal a27_a26_a25 : std_logic;
    signal a26_a25_a24 : std_logic;
    signal a25_a24_a23 : std_logic;
    signal a24_a23_a22 : std_logic;
    signal a23_a22_a21 : std_logic;
    signal a22_a21_a20 : std_logic;
    signal a21_a20_a19 : std_logic;
    signal a20_a19_a18 : std_logic;
    signal a19_a18_a17 : std_logic;
    signal a18_a17_a16 : std_logic;
    signal a17_a16_a15 : std_logic;
    signal a16_a15_a14 : std_logic;
    signal a15_a14_a13 : std_logic;
    signal a14_a13_a12 : std_logic;
    signal a13_a12_a11 : std_logic;
    signal a12_a11_a10 : std_logic;
    signal a11_a10_a9 : std_logic;
    signal a10_a9_a8 : std_logic;
    signal a9_a8_a7 : std_logic;
    signal a8_a7_a6 : std_logic;
    signal a7_a6_a5 : std_logic;
    signal a6_a5_a4 : std_logic;
    signal a5_a4_a3 : std_logic;
    signal a4_a3_a2 : std_logic;
    signal a3_a2_a1 : std_logic;
    signal a31_a23_a22_a21 : std_logic;
    signal a31_a30_a22_a21_a20 : std_logic;
    signal a31_a30_a29_a21_a20_a19 : std_logic;
    signal a30_a29_a28_a20_a19_a18 : std_logic;
    signal a29_a28_a27_a19_a18_a17 : std_logic;
    signal a28_a27_a26_a18_a17_a16 : std_logic;
    signal a27_a26_a25_a17_a16_a15 : std_logic;
    signal a26_a25_a24_a16_a15_a14 : std_logic;
    signal a25_a24_a23_a15_a14_a13 : std_logic;
    signal a24_a23_a22_a14_a13_a12 : std_logic;
    signal a31_a23_a22_a21_a13_a12_a11 : std_logic;
    signal a31_a30_a22_a21_a20_a12_a11_a10 : std_logic;
    signal a31_a30_a29_a21_a20_a19_a11_a10_a9 : std_logic;
    signal a30_a29_a28_a20_a19_a18_a10_a9_a8 : std_logic;
    signal a29_a28_a27_a19_a18_a17_a9_a8_a7 : std_logic;
    signal a28_a27_a26_a18_a17_a16_a8_a7_a6 : std_logic;
    signal a27_a26_a25_a17_a16_a15_a7_a6_a5 : std_logic;
    signal a26_a25_a24_a16_a15_a14_a6_a5_a4 : std_logic;
    signal a25_a24_a23_a15_a14_a13_a5_a4_a3 : std_logic;
    signal a24_a23_a22_a14_a13_a12_a4_a3_a2 : std_logic;
    signal a23_a22_a21_a13_a12_a11_a3_a2_a1 : std_logic;
    signal a31_a20_a10_a0 : std_logic;
    signal a31_a30_a19_a9 : std_logic;
    signal a30_a29_a18_a8 : std_logic;
    signal a29_a28_a17_a7 : std_logic;
    signal a28_a27_a16_a6 : std_logic;
    signal a27_a26_a15_a5 : std_logic;
    signal a26_a25_a14_a4 : std_logic;
    signal a25_a24_a13_a3 : std_logic;
    signal a24_a23_a12_a2 : std_logic;
    signal a31_a23_a22_a11_a1 : std_logic;
begin
    -- "f-network"
    a31_a30 <= a31 xor a30;
    a30_a29 <= a30 xor a29;
    a29_a28 <= a29 xor a28;
    a28_a27 <= a28 xor a27;
    a27_a26 <= a27 xor a26;
    a26_a25 <= a26 xor a25;
    a25_a24 <= a25 xor a24;
    a24_a23 <= a24 xor a23;
    a23_a22 <= a23 xor a22;
    a22_a21 <= a22 xor a21;
    a21_a20 <= a21 xor a20;
    a20_a19 <= a20 xor a19;
    a19_a18 <= a19 xor a18;
    a18_a17 <= a18 xor a17;
    a17_a16 <= a17 xor a16;
    a16_a15 <= a16 xor a15;
    a15_a14 <= a15 xor a14;
    a14_a13 <= a14 xor a13;
    a13_a12 <= a13 xor a12;
    a12_a11 <= a12 xor a11;
    a11_a10 <= a11 xor a10;
    a10_a9 <= a10 xor a9;
    a9_a8 <= a9 xor a8;
    a8_a7 <= a8 xor a7;
    a7_a6 <= a7 xor a6;
    a6_a5 <= a6 xor a5;
    a5_a4 <= a5 xor a4;
    a4_a3 <= a4 xor a3;
    a3_a2 <= a3 xor a2;
    a2_a1 <= a2 xor a1;
    -- a1_a0 <= a1 xor a0;  -- never used

    a31_a21 <= a31 xor a21;
    a30_a20 <= a30 xor a20;
    a29_a19 <= a29 xor a19;
    a28_a18 <= a28 xor a18;
    a27_a17 <= a27 xor a17;
    a26_a16 <= a26 xor a16;
    a25_a15 <= a25 xor a15;
    a24_a14 <= a24 xor a14;
    a23_a13 <= a23 xor a13;
    a22_a12 <= a22 xor a12;
    a21_a11 <= a21 xor a11;
    --a20_a10 <= a20 xor a10;  -- never used
    a19_a9 <= a19 xor a9;
    a18_a8 <= a18 xor a8;
    a17_a7 <= a17 xor a7;
    a16_a6 <= a16 xor a6;
    a15_a5 <= a15 xor a5;
    a14_a4 <= a14 xor a4;
    a13_a3 <= a13 xor a3;
    a12_a2 <= a12 xor a2;
    a11_a1 <= a11 xor a1;
    a10_a0 <= a10 xor a0;

    a31_a1 <= a31 xor a1;

    a31_a0 <= a31 xor a0;

    a31_a30_a29 <= a31_a30 xor a29;
    a30_a29_a28 <= a30_a29 xor a28;
    a29_a28_a27 <= a29_a28 xor a27;
    a28_a27_a26 <= a28_a27 xor a26;
    a27_a26_a25 <= a27_a26 xor a25;
    a26_a25_a24 <= a26_a25 xor a24;
    a25_a24_a23 <= a25_a24 xor a23;
    a24_a23_a22 <= a24_a23 xor a22;
    a23_a22_a21 <= a23_a22 xor a21;
    a22_a21_a20 <= a22_a21 xor a20;
    a21_a20_a19 <= a21_a20 xor a19;
    a20_a19_a18 <= a20_a19 xor a18;
    a19_a18_a17 <= a19_a18 xor a17;
    a18_a17_a16 <= a18_a17 xor a16;
    a17_a16_a15 <= a17_a16 xor a15;
    a16_a15_a14 <= a16_a15 xor a14;
    a15_a14_a13 <= a15_a14 xor a13;
    a14_a13_a12 <= a14_a13 xor a12;
    a13_a12_a11 <= a13_a12 xor a11;
    a12_a11_a10 <= a12_a11 xor a10;
    a11_a10_a9 <= a11_a10 xor a9;
    a10_a9_a8 <= a10_a9 xor a8;
    a9_a8_a7 <= a9_a8 xor a7;
    a8_a7_a6 <= a8_a7 xor a6;
    a7_a6_a5 <= a7_a6 xor a5;
    a6_a5_a4 <= a6_a5 xor a4;
    a5_a4_a3 <= a5_a4 xor a3;
    a4_a3_a2 <= a4_a3 xor a2;
    a3_a2_a1 <= a3_a2 xor a1;

    a31_a22_a21 <= a31 xor a22_a21;

    a31_a21_a11 <= a31_a21 xor a11;
    a30_a20_a10 <= a30_a20 xor a10;
    a29_a19_a9 <= a29_a19 xor a9;
    a28_a18_a8 <= a28_a18 xor a8;
    a27_a17_a7 <= a27_a17 xor a7;
    a26_a16_a6 <= a26_a16 xor a6;
    a25_a15_a5 <= a25_a15 xor a5;
    a24_a14_a4 <= a24_a14 xor a4;
    a23_a13_a3 <= a23_a13 xor a3;
    a22_a12_a2 <= a22_a12 xor a2;
    a21_a11_a1 <= a21_a11 xor a1;

    a31_a30_a0 <= a31_a30 xor a0;

    a31_a30_a21_a20 <= a31_a30 xor a21_a20;
    a30_a29_a20_a19 <= a30_a29 xor a20_a19;
    a29_a28_a19_a18 <= a29_a28 xor a19_a18;
    a28_a27_a18_a17 <= a28_a27 xor a18_a17;
    a27_a26_a17_a16 <= a27_a26 xor a17_a16;
    a26_a25_a16_a15 <= a26_a25 xor a16_a15;
    a25_a24_a15_a14 <= a25_a24 xor a15_a14;
    a24_a23_a14_a13 <= a24_a23 xor a14_a13;
    a23_a22_a13_a12 <= a23_a22 xor a13_a12;

    a31_a23_a22_a21 <= a31 xor a23_a22_a21;

    a31_a20_a10_a0 <= a31 xor a20 xor a10_a0;

    a31_a30_a19_a9 <= a31_a30 xor a19_a9;
    a30_a29_a18_a8 <= a30_a29 xor a18_a8;
    a29_a28_a17_a7 <= a29_a28 xor a17_a7;
    a28_a27_a16_a6 <= a28_a27 xor a16_a6;
    a27_a26_a15_a5 <= a27_a26 xor a15_a5;
    a26_a25_a14_a4 <= a26_a25 xor a14_a4;
    a25_a24_a13_a3 <= a25_a24 xor a13_a3;
    a24_a23_a12_a2 <= a24_a23 xor a12_a2;

    a31_a23_a22_a11_a1 <= a31 xor a23_a22 xor a11_a1;
    a31_a30_a22_a21_a20 <= a31_a30 xor a22_a21_a20;

    a31_a22_a21_a12_a11 <= a31 xor a22_a21 xor a12_a11;

    a31_a30_a21_a20_a11_a10 <= a31_a30 xor a21_a20 xor a11_a10;
    a30_a29_a20_a19_a10_a9 <= a30_a29 xor a20_a19 xor a10_a9;
    a29_a28_a19_a18_a9_a8 <= a29_a28 xor a19_a18 xor a9_a8;
    a28_a27_a18_a17_a8_a7 <= a28_a27 xor a18_a17 xor a8_a7;
    a27_a26_a17_a16_a7_a6 <= a27_a26 xor a17_a16 xor a7_a6;
    a26_a25_a16_a15_a6_a5 <= a26_a25 xor a16_a15 xor a6_a5;
    a25_a24_a15_a14_a5_a4 <= a25_a24 xor a15_a14 xor a5_a4;
    a24_a23_a14_a13_a4_a3 <= a24_a23 xor a14_a13 xor a4_a3;
    a23_a22_a13_a12_a3_a2 <= a23_a22 xor a13_a12 xor a3_a2;
    a22_a21_a12_a11_a2_a1 <= a22_a21 xor a12_a11 xor a2_a1;

    a31_a30_a29_a21_a20_a19 <= a31_a30_a29 xor a21_a20_a19;
    a30_a29_a28_a20_a19_a18 <= a30_a29_a28 xor a20_a19_a18;
    a29_a28_a27_a19_a18_a17 <= a29_a28_a27 xor a19_a18_a17;
    a28_a27_a26_a18_a17_a16 <= a28_a27_a26 xor a18_a17_a16;
    a27_a26_a25_a17_a16_a15 <= a27_a26_a25 xor a17_a16_a15;
    a26_a25_a24_a16_a15_a14 <= a26_a25_a24 xor a16_a15_a14;
    a25_a24_a23_a15_a14_a13 <= a25_a24_a23 xor a15_a14_a13;
    a24_a23_a22_a14_a13_a12 <= a24_a23_a22 xor a14_a13_a12;

    a31_a23_a22_a21_a13_a12_a11 <= a31 xor a23_a22_a21 xor a13_a12_a11;

    a31_a30_a22_a21_a20_a12_a11_a10 <= a31_a30 xor a22_a21_a20 xor a12_a11_a10;

    a31_a30_a29_a21_a20_a19_a11_a10_a9 <= a31_a30_a29 xor a21_a20_a19 xor a11_a10_a9;
    a30_a29_a28_a20_a19_a18_a10_a9_a8 <= a30_a29_a28 xor a20_a19_a18 xor a10_a9_a8;
    a29_a28_a27_a19_a18_a17_a9_a8_a7 <= a29_a28_a27 xor a19_a18_a17 xor a9_a8_a7;
    a28_a27_a26_a18_a17_a16_a8_a7_a6 <= a28_a27_a26 xor a18_a17_a16 xor a8_a7_a6;
    a27_a26_a25_a17_a16_a15_a7_a6_a5 <= a27_a26_a25 xor a17_a16_a15 xor a7_a6_a5;
    a26_a25_a24_a16_a15_a14_a6_a5_a4 <= a26_a25_a24 xor a16_a15_a14 xor a6_a5_a4;
    a25_a24_a23_a15_a14_a13_a5_a4_a3 <= a25_a24_a23 xor a15_a14_a13 xor a5_a4_a3;
    a24_a23_a22_a14_a13_a12_a4_a3_a2 <= a24_a23_a22 xor a14_a13_a12 xor a4_a3_a2;
    a23_a22_a21_a13_a12_a11_a3_a2_a1 <= a23_a22_a21 xor a13_a12_a11 xor a3_a2_a1;

    -- "IP-network"
    ab31 <= ((((((a31_a20_a10_a0 and b31) xor (a21_a11_a1 and b30)) xor ((a22_a12_a2 and b29) xor (a23_a13_a3 and b28))) xor (((a24_a14_a4 and b27) xor (a25_a15_a5 and b26)) xor ((a26_a16_a6 and b25) xor (a27_a17_a7 and b24)))) xor ((((a28_a18_a8 and b23) xor (a29_a19_a9 and b22)) xor ((a30_a20_a10 and b21) xor (a31_a21_a11 and b20))) xor (((a22_a12 and b19) xor (a23_a13 and b18)) xor ((a24_a14 and b17) xor (a25_a15 and b16))))) xor (((((a26_a16 and b15) xor (a27_a17 and b14)) xor ((a28_a18 and b13) xor (a29_a19 and b12))) xor (((a30_a20 and b11) xor (a31_a21 and b10)) xor ((a22 and b9) xor (a23 and b8)))) xor ((((a24 and b7) xor (a25 and b6)) xor ((a26 and b5) xor (a27 and b4))) xor (((a28 and b3) xor (a29 and b2)) xor ((a30 and b1) xor (a31 and b0))))));
    ab30 <= ((((((a31_a30_a19_a9 and b31) xor (a31_a20_a10_a0 and b30)) xor ((a21_a11_a1 and b29) xor (a22_a12_a2 and b28))) xor (((a23_a13_a3 and b27) xor (a24_a14_a4 and b26)) xor ((a25_a15_a5 and b25) xor (a26_a16_a6 and b24)))) xor ((((a27_a17_a7 and b23) xor (a28_a18_a8 and b22)) xor ((a29_a19_a9 and b21) xor (a30_a20_a10 and b20))) xor (((a31_a21_a11 and b19) xor (a22_a12 and b18)) xor ((a23_a13 and b17) xor (a24_a14 and b16))))) xor (((((a25_a15 and b15) xor (a26_a16 and b14)) xor ((a27_a17 and b13) xor (a28_a18 and b12))) xor (((a29_a19 and b11) xor (a30_a20 and b10)) xor ((a31_a21 and b9) xor (a22 and b8)))) xor ((((a23 and b7) xor (a24 and b6)) xor ((a25 and b5) xor (a26 and b4))) xor (((a27 and b3) xor (a28 and b2)) xor ((a29 and b1) xor (a30 and b0))))));
    ab29 <= ((((((a30_a29_a18_a8 and b31) xor (a31_a30_a19_a9 and b30)) xor ((a31_a20_a10_a0 and b29) xor (a21_a11_a1 and b28))) xor (((a22_a12_a2 and b27) xor (a23_a13_a3 and b26)) xor ((a24_a14_a4 and b25) xor (a25_a15_a5 and b24)))) xor ((((a26_a16_a6 and b23) xor (a27_a17_a7 and b22)) xor ((a28_a18_a8 and b21) xor (a29_a19_a9 and b20))) xor (((a30_a20_a10 and b19) xor (a31_a21_a11 and b18)) xor ((a22_a12 and b17) xor (a23_a13 and b16))))) xor (((((a24_a14 and b15) xor (a25_a15 and b14)) xor ((a26_a16 and b13) xor (a27_a17 and b12))) xor (((a28_a18 and b11) xor (a29_a19 and b10)) xor ((a30_a20 and b9) xor (a31_a21 and b8)))) xor ((((a22 and b7) xor (a23 and b6)) xor ((a24 and b5) xor (a25 and b4))) xor (((a26 and b3) xor (a27 and b2)) xor ((a28 and b1) xor (a29 and b0))))));
    ab28 <= ((((((a29_a28_a17_a7 and b31) xor (a30_a29_a18_a8 and b30)) xor ((a31_a30_a19_a9 and b29) xor (a31_a20_a10_a0 and b28))) xor (((a21_a11_a1 and b27) xor (a22_a12_a2 and b26)) xor ((a23_a13_a3 and b25) xor (a24_a14_a4 and b24)))) xor ((((a25_a15_a5 and b23) xor (a26_a16_a6 and b22)) xor ((a27_a17_a7 and b21) xor (a28_a18_a8 and b20))) xor (((a29_a19_a9 and b19) xor (a30_a20_a10 and b18)) xor ((a31_a21_a11 and b17) xor (a22_a12 and b16))))) xor (((((a23_a13 and b15) xor (a24_a14 and b14)) xor ((a25_a15 and b13) xor (a26_a16 and b12))) xor (((a27_a17 and b11) xor (a28_a18 and b10)) xor ((a29_a19 and b9) xor (a30_a20 and b8)))) xor ((((a31_a21 and b7) xor (a22 and b6)) xor ((a23 and b5) xor (a24 and b4))) xor (((a25 and b3) xor (a26 and b2)) xor ((a27 and b1) xor (a28 and b0))))));
    ab27 <= ((((((a28_a27_a16_a6 and b31) xor (a29_a28_a17_a7 and b30)) xor ((a30_a29_a18_a8 and b29) xor (a31_a30_a19_a9 and b28))) xor (((a31_a20_a10_a0 and b27) xor (a21_a11_a1 and b26)) xor ((a22_a12_a2 and b25) xor (a23_a13_a3 and b24)))) xor ((((a24_a14_a4 and b23) xor (a25_a15_a5 and b22)) xor ((a26_a16_a6 and b21) xor (a27_a17_a7 and b20))) xor (((a28_a18_a8 and b19) xor (a29_a19_a9 and b18)) xor ((a30_a20_a10 and b17) xor (a31_a21_a11 and b16))))) xor (((((a22_a12 and b15) xor (a23_a13 and b14)) xor ((a24_a14 and b13) xor (a25_a15 and b12))) xor (((a26_a16 and b11) xor (a27_a17 and b10)) xor ((a28_a18 and b9) xor (a29_a19 and b8)))) xor ((((a30_a20 and b7) xor (a31_a21 and b6)) xor ((a22 and b5) xor (a23 and b4))) xor (((a24 and b3) xor (a25 and b2)) xor ((a26 and b1) xor (a27 and b0))))));
    ab26 <= ((((((a27_a26_a15_a5 and b31) xor (a28_a27_a16_a6 and b30)) xor ((a29_a28_a17_a7 and b29) xor (a30_a29_a18_a8 and b28))) xor (((a31_a30_a19_a9 and b27) xor (a31_a20_a10_a0 and b26)) xor ((a21_a11_a1 and b25) xor (a22_a12_a2 and b24)))) xor ((((a23_a13_a3 and b23) xor (a24_a14_a4 and b22)) xor ((a25_a15_a5 and b21) xor (a26_a16_a6 and b20))) xor (((a27_a17_a7 and b19) xor (a28_a18_a8 and b18)) xor ((a29_a19_a9 and b17) xor (a30_a20_a10 and b16))))) xor (((((a31_a21_a11 and b15) xor (a22_a12 and b14)) xor ((a23_a13 and b13) xor (a24_a14 and b12))) xor (((a25_a15 and b11) xor (a26_a16 and b10)) xor ((a27_a17 and b9) xor (a28_a18 and b8)))) xor ((((a29_a19 and b7) xor (a30_a20 and b6)) xor ((a31_a21 and b5) xor (a22 and b4))) xor (((a23 and b3) xor (a24 and b2)) xor ((a25 and b1) xor (a26 and b0))))));
    ab25 <= ((((((a26_a25_a14_a4 and b31) xor (a27_a26_a15_a5 and b30)) xor ((a28_a27_a16_a6 and b29) xor (a29_a28_a17_a7 and b28))) xor (((a30_a29_a18_a8 and b27) xor (a31_a30_a19_a9 and b26)) xor ((a31_a20_a10_a0 and b25) xor (a21_a11_a1 and b24)))) xor ((((a22_a12_a2 and b23) xor (a23_a13_a3 and b22)) xor ((a24_a14_a4 and b21) xor (a25_a15_a5 and b20))) xor (((a26_a16_a6 and b19) xor (a27_a17_a7 and b18)) xor ((a28_a18_a8 and b17) xor (a29_a19_a9 and b16))))) xor (((((a30_a20_a10 and b15) xor (a31_a21_a11 and b14)) xor ((a22_a12 and b13) xor (a23_a13 and b12))) xor (((a24_a14 and b11) xor (a25_a15 and b10)) xor ((a26_a16 and b9) xor (a27_a17 and b8)))) xor ((((a28_a18 and b7) xor (a29_a19 and b6)) xor ((a30_a20 and b5) xor (a31_a21 and b4))) xor (((a22 and b3) xor (a23 and b2)) xor ((a24 and b1) xor (a25 and b0))))));
    ab24 <= ((((((a25_a24_a13_a3 and b31) xor (a26_a25_a14_a4 and b30)) xor ((a27_a26_a15_a5 and b29) xor (a28_a27_a16_a6 and b28))) xor (((a29_a28_a17_a7 and b27) xor (a30_a29_a18_a8 and b26)) xor ((a31_a30_a19_a9 and b25) xor (a31_a20_a10_a0 and b24)))) xor ((((a21_a11_a1 and b23) xor (a22_a12_a2 and b22)) xor ((a23_a13_a3 and b21) xor (a24_a14_a4 and b20))) xor (((a25_a15_a5 and b19) xor (a26_a16_a6 and b18)) xor ((a27_a17_a7 and b17) xor (a28_a18_a8 and b16))))) xor (((((a29_a19_a9 and b15) xor (a30_a20_a10 and b14)) xor ((a31_a21_a11 and b13) xor (a22_a12 and b12))) xor (((a23_a13 and b11) xor (a24_a14 and b10)) xor ((a25_a15 and b9) xor (a26_a16 and b8)))) xor ((((a27_a17 and b7) xor (a28_a18 and b6)) xor ((a29_a19 and b5) xor (a30_a20 and b4))) xor (((a31_a21 and b3) xor (a22 and b2)) xor ((a23 and b1) xor (a24 and b0))))));
    ab23 <= ((((((a24_a23_a12_a2 and b31) xor (a25_a24_a13_a3 and b30)) xor ((a26_a25_a14_a4 and b29) xor (a27_a26_a15_a5 and b28))) xor (((a28_a27_a16_a6 and b27) xor (a29_a28_a17_a7 and b26)) xor ((a30_a29_a18_a8 and b25) xor (a31_a30_a19_a9 and b24)))) xor ((((a31_a20_a10_a0 and b23) xor (a21_a11_a1 and b22)) xor ((a22_a12_a2 and b21) xor (a23_a13_a3 and b20))) xor (((a24_a14_a4 and b19) xor (a25_a15_a5 and b18)) xor ((a26_a16_a6 and b17) xor (a27_a17_a7 and b16))))) xor (((((a28_a18_a8 and b15) xor (a29_a19_a9 and b14)) xor ((a30_a20_a10 and b13) xor (a31_a21_a11 and b12))) xor (((a22_a12 and b11) xor (a23_a13 and b10)) xor ((a24_a14 and b9) xor (a25_a15 and b8)))) xor ((((a26_a16 and b7) xor (a27_a17 and b6)) xor ((a28_a18 and b5) xor (a29_a19 and b4))) xor (((a30_a20 and b3) xor (a31_a21 and b2)) xor ((a22 and b1) xor (a23 and b0))))));
    ab22 <= ((((((a31_a23_a22_a11_a1 and b31) xor (a24_a23_a12_a2 and b30)) xor ((a25_a24_a13_a3 and b29) xor (a26_a25_a14_a4 and b28))) xor (((a27_a26_a15_a5 and b27) xor (a28_a27_a16_a6 and b26)) xor ((a29_a28_a17_a7 and b25) xor (a30_a29_a18_a8 and b24)))) xor ((((a31_a30_a19_a9 and b23) xor (a31_a20_a10_a0 and b22)) xor ((a21_a11_a1 and b21) xor (a22_a12_a2 and b20))) xor (((a23_a13_a3 and b19) xor (a24_a14_a4 and b18)) xor ((a25_a15_a5 and b17) xor (a26_a16_a6 and b16))))) xor (((((a27_a17_a7 and b15) xor (a28_a18_a8 and b14)) xor ((a29_a19_a9 and b13) xor (a30_a20_a10 and b12))) xor (((a31_a21_a11 and b11) xor (a22_a12 and b10)) xor ((a23_a13 and b9) xor (a24_a14 and b8)))) xor ((((a25_a15 and b7) xor (a26_a16 and b6)) xor ((a27_a17 and b5) xor (a28_a18 and b4))) xor (((a29_a19 and b3) xor (a30_a20 and b2)) xor ((a31_a21 and b1) xor (a22 and b0))))));
    ab21 <= ((((((a31_a30_a22_a21_a20 and b31) xor (a31_a23_a22_a21 and b30)) xor ((a24_a23_a22 and b29) xor (a25_a24_a23 and b28))) xor (((a26_a25_a24 and b27) xor (a27_a26_a25 and b26)) xor ((a28_a27_a26 and b25) xor (a29_a28_a27 and b24)))) xor ((((a30_a29_a28 and b23) xor (a31_a30_a29 and b22)) xor ((a31_a30_a0 and b21) xor (a31_a1 and b20))) xor (((a2 and b19) xor (a3 and b18)) xor ((a4 and b17) xor (a5 and b16))))) xor (((((a6 and b15) xor (a7 and b14)) xor ((a8 and b13) xor (a9 and b12))) xor (((a10 and b11) xor (a11 and b10)) xor ((a12 and b9) xor (a13 and b8)))) xor ((((a14 and b7) xor (a15 and b6)) xor ((a16 and b5) xor (a17 and b4))) xor (((a18 and b3) xor (a19 and b2)) xor ((a20 and b1) xor (a21 and b0))))));
    ab20 <= ((((((a31_a30_a29_a21_a20_a19 and b31) xor (a31_a30_a22_a21_a20 and b30)) xor ((a31_a23_a22_a21 and b29) xor (a24_a23_a22 and b28))) xor (((a25_a24_a23 and b27) xor (a26_a25_a24 and b26)) xor ((a27_a26_a25 and b25) xor (a28_a27_a26 and b24)))) xor ((((a29_a28_a27 and b23) xor (a30_a29_a28 and b22)) xor ((a31_a30_a29 and b21) xor (a31_a30_a0 and b20))) xor (((a31_a1 and b19) xor (a2 and b18)) xor ((a3 and b17) xor (a4 and b16))))) xor (((((a5 and b15) xor (a6 and b14)) xor ((a7 and b13) xor (a8 and b12))) xor (((a9 and b11) xor (a10 and b10)) xor ((a11 and b9) xor (a12 and b8)))) xor ((((a13 and b7) xor (a14 and b6)) xor ((a15 and b5) xor (a16 and b4))) xor (((a17 and b3) xor (a18 and b2)) xor ((a19 and b1) xor (a20 and b0))))));
    ab19 <= ((((((a30_a29_a28_a20_a19_a18 and b31) xor (a31_a30_a29_a21_a20_a19 and b30)) xor ((a31_a30_a22_a21_a20 and b29) xor (a31_a23_a22_a21 and b28))) xor (((a24_a23_a22 and b27) xor (a25_a24_a23 and b26)) xor ((a26_a25_a24 and b25) xor (a27_a26_a25 and b24)))) xor ((((a28_a27_a26 and b23) xor (a29_a28_a27 and b22)) xor ((a30_a29_a28 and b21) xor (a31_a30_a29 and b20))) xor (((a31_a30_a0 and b19) xor (a31_a1 and b18)) xor ((a2 and b17) xor (a3 and b16))))) xor (((((a4 and b15) xor (a5 and b14)) xor ((a6 and b13) xor (a7 and b12))) xor (((a8 and b11) xor (a9 and b10)) xor ((a10 and b9) xor (a11 and b8)))) xor ((((a12 and b7) xor (a13 and b6)) xor ((a14 and b5) xor (a15 and b4))) xor (((a16 and b3) xor (a17 and b2)) xor ((a18 and b1) xor (a19 and b0))))));
    ab18 <= ((((((a29_a28_a27_a19_a18_a17 and b31) xor (a30_a29_a28_a20_a19_a18 and b30)) xor ((a31_a30_a29_a21_a20_a19 and b29) xor (a31_a30_a22_a21_a20 and b28))) xor (((a31_a23_a22_a21 and b27) xor (a24_a23_a22 and b26)) xor ((a25_a24_a23 and b25) xor (a26_a25_a24 and b24)))) xor ((((a27_a26_a25 and b23) xor (a28_a27_a26 and b22)) xor ((a29_a28_a27 and b21) xor (a30_a29_a28 and b20))) xor (((a31_a30_a29 and b19) xor (a31_a30_a0 and b18)) xor ((a31_a1 and b17) xor (a2 and b16))))) xor (((((a3 and b15) xor (a4 and b14)) xor ((a5 and b13) xor (a6 and b12))) xor (((a7 and b11) xor (a8 and b10)) xor ((a9 and b9) xor (a10 and b8)))) xor ((((a11 and b7) xor (a12 and b6)) xor ((a13 and b5) xor (a14 and b4))) xor (((a15 and b3) xor (a16 and b2)) xor ((a17 and b1) xor (a18 and b0))))));
    ab17 <= ((((((a28_a27_a26_a18_a17_a16 and b31) xor (a29_a28_a27_a19_a18_a17 and b30)) xor ((a30_a29_a28_a20_a19_a18 and b29) xor (a31_a30_a29_a21_a20_a19 and b28))) xor (((a31_a30_a22_a21_a20 and b27) xor (a31_a23_a22_a21 and b26)) xor ((a24_a23_a22 and b25) xor (a25_a24_a23 and b24)))) xor ((((a26_a25_a24 and b23) xor (a27_a26_a25 and b22)) xor ((a28_a27_a26 and b21) xor (a29_a28_a27 and b20))) xor (((a30_a29_a28 and b19) xor (a31_a30_a29 and b18)) xor ((a31_a30_a0 and b17) xor (a31_a1 and b16))))) xor (((((a2 and b15) xor (a3 and b14)) xor ((a4 and b13) xor (a5 and b12))) xor (((a6 and b11) xor (a7 and b10)) xor ((a8 and b9) xor (a9 and b8)))) xor ((((a10 and b7) xor (a11 and b6)) xor ((a12 and b5) xor (a13 and b4))) xor (((a14 and b3) xor (a15 and b2)) xor ((a16 and b1) xor (a17 and b0))))));
    ab16 <= ((((((a27_a26_a25_a17_a16_a15 and b31) xor (a28_a27_a26_a18_a17_a16 and b30)) xor ((a29_a28_a27_a19_a18_a17 and b29) xor (a30_a29_a28_a20_a19_a18 and b28))) xor (((a31_a30_a29_a21_a20_a19 and b27) xor (a31_a30_a22_a21_a20 and b26)) xor ((a31_a23_a22_a21 and b25) xor (a24_a23_a22 and b24)))) xor ((((a25_a24_a23 and b23) xor (a26_a25_a24 and b22)) xor ((a27_a26_a25 and b21) xor (a28_a27_a26 and b20))) xor (((a29_a28_a27 and b19) xor (a30_a29_a28 and b18)) xor ((a31_a30_a29 and b17) xor (a31_a30_a0 and b16))))) xor (((((a31_a1 and b15) xor (a2 and b14)) xor ((a3 and b13) xor (a4 and b12))) xor (((a5 and b11) xor (a6 and b10)) xor ((a7 and b9) xor (a8 and b8)))) xor ((((a9 and b7) xor (a10 and b6)) xor ((a11 and b5) xor (a12 and b4))) xor (((a13 and b3) xor (a14 and b2)) xor ((a15 and b1) xor (a16 and b0))))));
    ab15 <= ((((((a26_a25_a24_a16_a15_a14 and b31) xor (a27_a26_a25_a17_a16_a15 and b30)) xor ((a28_a27_a26_a18_a17_a16 and b29) xor (a29_a28_a27_a19_a18_a17 and b28))) xor (((a30_a29_a28_a20_a19_a18 and b27) xor (a31_a30_a29_a21_a20_a19 and b26)) xor ((a31_a30_a22_a21_a20 and b25) xor (a31_a23_a22_a21 and b24)))) xor ((((a24_a23_a22 and b23) xor (a25_a24_a23 and b22)) xor ((a26_a25_a24 and b21) xor (a27_a26_a25 and b20))) xor (((a28_a27_a26 and b19) xor (a29_a28_a27 and b18)) xor ((a30_a29_a28 and b17) xor (a31_a30_a29 and b16))))) xor (((((a31_a30_a0 and b15) xor (a31_a1 and b14)) xor ((a2 and b13) xor (a3 and b12))) xor (((a4 and b11) xor (a5 and b10)) xor ((a6 and b9) xor (a7 and b8)))) xor ((((a8 and b7) xor (a9 and b6)) xor ((a10 and b5) xor (a11 and b4))) xor (((a12 and b3) xor (a13 and b2)) xor ((a14 and b1) xor (a15 and b0))))));
    ab14 <= ((((((a25_a24_a23_a15_a14_a13 and b31) xor (a26_a25_a24_a16_a15_a14 and b30)) xor ((a27_a26_a25_a17_a16_a15 and b29) xor (a28_a27_a26_a18_a17_a16 and b28))) xor (((a29_a28_a27_a19_a18_a17 and b27) xor (a30_a29_a28_a20_a19_a18 and b26)) xor ((a31_a30_a29_a21_a20_a19 and b25) xor (a31_a30_a22_a21_a20 and b24)))) xor ((((a31_a23_a22_a21 and b23) xor (a24_a23_a22 and b22)) xor ((a25_a24_a23 and b21) xor (a26_a25_a24 and b20))) xor (((a27_a26_a25 and b19) xor (a28_a27_a26 and b18)) xor ((a29_a28_a27 and b17) xor (a30_a29_a28 and b16))))) xor (((((a31_a30_a29 and b15) xor (a31_a30_a0 and b14)) xor ((a31_a1 and b13) xor (a2 and b12))) xor (((a3 and b11) xor (a4 and b10)) xor ((a5 and b9) xor (a6 and b8)))) xor ((((a7 and b7) xor (a8 and b6)) xor ((a9 and b5) xor (a10 and b4))) xor (((a11 and b3) xor (a12 and b2)) xor ((a13 and b1) xor (a14 and b0))))));
    ab13 <= ((((((a24_a23_a22_a14_a13_a12 and b31) xor (a25_a24_a23_a15_a14_a13 and b30)) xor ((a26_a25_a24_a16_a15_a14 and b29) xor (a27_a26_a25_a17_a16_a15 and b28))) xor (((a28_a27_a26_a18_a17_a16 and b27) xor (a29_a28_a27_a19_a18_a17 and b26)) xor ((a30_a29_a28_a20_a19_a18 and b25) xor (a31_a30_a29_a21_a20_a19 and b24)))) xor ((((a31_a30_a22_a21_a20 and b23) xor (a31_a23_a22_a21 and b22)) xor ((a24_a23_a22 and b21) xor (a25_a24_a23 and b20))) xor (((a26_a25_a24 and b19) xor (a27_a26_a25 and b18)) xor ((a28_a27_a26 and b17) xor (a29_a28_a27 and b16))))) xor (((((a30_a29_a28 and b15) xor (a31_a30_a29 and b14)) xor ((a31_a30_a0 and b13) xor (a31_a1 and b12))) xor (((a2 and b11) xor (a3 and b10)) xor ((a4 and b9) xor (a5 and b8)))) xor ((((a6 and b7) xor (a7 and b6)) xor ((a8 and b5) xor (a9 and b4))) xor (((a10 and b3) xor (a11 and b2)) xor ((a12 and b1) xor (a13 and b0))))));
    ab12 <= ((((((a31_a23_a22_a21_a13_a12_a11 and b31) xor (a24_a23_a22_a14_a13_a12 and b30)) xor ((a25_a24_a23_a15_a14_a13 and b29) xor (a26_a25_a24_a16_a15_a14 and b28))) xor (((a27_a26_a25_a17_a16_a15 and b27) xor (a28_a27_a26_a18_a17_a16 and b26)) xor ((a29_a28_a27_a19_a18_a17 and b25) xor (a30_a29_a28_a20_a19_a18 and b24)))) xor ((((a31_a30_a29_a21_a20_a19 and b23) xor (a31_a30_a22_a21_a20 and b22)) xor ((a31_a23_a22_a21 and b21) xor (a24_a23_a22 and b20))) xor (((a25_a24_a23 and b19) xor (a26_a25_a24 and b18)) xor ((a27_a26_a25 and b17) xor (a28_a27_a26 and b16))))) xor (((((a29_a28_a27 and b15) xor (a30_a29_a28 and b14)) xor ((a31_a30_a29 and b13) xor (a31_a30_a0 and b12))) xor (((a31_a1 and b11) xor (a2 and b10)) xor ((a3 and b9) xor (a4 and b8)))) xor ((((a5 and b7) xor (a6 and b6)) xor ((a7 and b5) xor (a8 and b4))) xor (((a9 and b3) xor (a10 and b2)) xor ((a11 and b1) xor (a12 and b0))))));
    ab11 <= ((((((a31_a30_a22_a21_a20_a12_a11_a10 and b31) xor (a31_a23_a22_a21_a13_a12_a11 and b30)) xor ((a24_a23_a22_a14_a13_a12 and b29) xor (a25_a24_a23_a15_a14_a13 and b28))) xor (((a26_a25_a24_a16_a15_a14 and b27) xor (a27_a26_a25_a17_a16_a15 and b26)) xor ((a28_a27_a26_a18_a17_a16 and b25) xor (a29_a28_a27_a19_a18_a17 and b24)))) xor ((((a30_a29_a28_a20_a19_a18 and b23) xor (a31_a30_a29_a21_a20_a19 and b22)) xor ((a31_a30_a22_a21_a20 and b21) xor (a31_a23_a22_a21 and b20))) xor (((a24_a23_a22 and b19) xor (a25_a24_a23 and b18)) xor ((a26_a25_a24 and b17) xor (a27_a26_a25 and b16))))) xor (((((a28_a27_a26 and b15) xor (a29_a28_a27 and b14)) xor ((a30_a29_a28 and b13) xor (a31_a30_a29 and b12))) xor (((a31_a30_a0 and b11) xor (a31_a1 and b10)) xor ((a2 and b9) xor (a3 and b8)))) xor ((((a4 and b7) xor (a5 and b6)) xor ((a6 and b5) xor (a7 and b4))) xor (((a8 and b3) xor (a9 and b2)) xor ((a10 and b1) xor (a11 and b0))))));
    ab10 <= ((((((a31_a30_a29_a21_a20_a19_a11_a10_a9 and b31) xor (a31_a30_a22_a21_a20_a12_a11_a10 and b30)) xor ((a31_a23_a22_a21_a13_a12_a11 and b29) xor (a24_a23_a22_a14_a13_a12 and b28))) xor (((a25_a24_a23_a15_a14_a13 and b27) xor (a26_a25_a24_a16_a15_a14 and b26)) xor ((a27_a26_a25_a17_a16_a15 and b25) xor (a28_a27_a26_a18_a17_a16 and b24)))) xor ((((a29_a28_a27_a19_a18_a17 and b23) xor (a30_a29_a28_a20_a19_a18 and b22)) xor ((a31_a30_a29_a21_a20_a19 and b21) xor (a31_a30_a22_a21_a20 and b20))) xor (((a31_a23_a22_a21 and b19) xor (a24_a23_a22 and b18)) xor ((a25_a24_a23 and b17) xor (a26_a25_a24 and b16))))) xor (((((a27_a26_a25 and b15) xor (a28_a27_a26 and b14)) xor ((a29_a28_a27 and b13) xor (a30_a29_a28 and b12))) xor (((a31_a30_a29 and b11) xor (a31_a30_a0 and b10)) xor ((a31_a1 and b9) xor (a2 and b8)))) xor ((((a3 and b7) xor (a4 and b6)) xor ((a5 and b5) xor (a6 and b4))) xor (((a7 and b3) xor (a8 and b2)) xor ((a9 and b1) xor (a10 and b0))))));
    ab9 <= ((((((a30_a29_a28_a20_a19_a18_a10_a9_a8 and b31) xor (a31_a30_a29_a21_a20_a19_a11_a10_a9 and b30)) xor ((a31_a30_a22_a21_a20_a12_a11_a10 and b29) xor (a31_a23_a22_a21_a13_a12_a11 and b28))) xor (((a24_a23_a22_a14_a13_a12 and b27) xor (a25_a24_a23_a15_a14_a13 and b26)) xor ((a26_a25_a24_a16_a15_a14 and b25) xor (a27_a26_a25_a17_a16_a15 and b24)))) xor ((((a28_a27_a26_a18_a17_a16 and b23) xor (a29_a28_a27_a19_a18_a17 and b22)) xor ((a30_a29_a28_a20_a19_a18 and b21) xor (a31_a30_a29_a21_a20_a19 and b20))) xor (((a31_a30_a22_a21_a20 and b19) xor (a31_a23_a22_a21 and b18)) xor ((a24_a23_a22 and b17) xor (a25_a24_a23 and b16))))) xor (((((a26_a25_a24 and b15) xor (a27_a26_a25 and b14)) xor ((a28_a27_a26 and b13) xor (a29_a28_a27 and b12))) xor (((a30_a29_a28 and b11) xor (a31_a30_a29 and b10)) xor ((a31_a30_a0 and b9) xor (a31_a1 and b8)))) xor ((((a2 and b7) xor (a3 and b6)) xor ((a4 and b5) xor (a5 and b4))) xor (((a6 and b3) xor (a7 and b2)) xor ((a8 and b1) xor (a9 and b0))))));
    ab8 <= ((((((a29_a28_a27_a19_a18_a17_a9_a8_a7 and b31) xor (a30_a29_a28_a20_a19_a18_a10_a9_a8 and b30)) xor ((a31_a30_a29_a21_a20_a19_a11_a10_a9 and b29) xor (a31_a30_a22_a21_a20_a12_a11_a10 and b28))) xor (((a31_a23_a22_a21_a13_a12_a11 and b27) xor (a24_a23_a22_a14_a13_a12 and b26)) xor ((a25_a24_a23_a15_a14_a13 and b25) xor (a26_a25_a24_a16_a15_a14 and b24)))) xor ((((a27_a26_a25_a17_a16_a15 and b23) xor (a28_a27_a26_a18_a17_a16 and b22)) xor ((a29_a28_a27_a19_a18_a17 and b21) xor (a30_a29_a28_a20_a19_a18 and b20))) xor (((a31_a30_a29_a21_a20_a19 and b19) xor (a31_a30_a22_a21_a20 and b18)) xor ((a31_a23_a22_a21 and b17) xor (a24_a23_a22 and b16))))) xor (((((a25_a24_a23 and b15) xor (a26_a25_a24 and b14)) xor ((a27_a26_a25 and b13) xor (a28_a27_a26 and b12))) xor (((a29_a28_a27 and b11) xor (a30_a29_a28 and b10)) xor ((a31_a30_a29 and b9) xor (a31_a30_a0 and b8)))) xor ((((a31_a1 and b7) xor (a2 and b6)) xor ((a3 and b5) xor (a4 and b4))) xor (((a5 and b3) xor (a6 and b2)) xor ((a7 and b1) xor (a8 and b0))))));
    ab7 <= ((((((a28_a27_a26_a18_a17_a16_a8_a7_a6 and b31) xor (a29_a28_a27_a19_a18_a17_a9_a8_a7 and b30)) xor ((a30_a29_a28_a20_a19_a18_a10_a9_a8 and b29) xor (a31_a30_a29_a21_a20_a19_a11_a10_a9 and b28))) xor (((a31_a30_a22_a21_a20_a12_a11_a10 and b27) xor (a31_a23_a22_a21_a13_a12_a11 and b26)) xor ((a24_a23_a22_a14_a13_a12 and b25) xor (a25_a24_a23_a15_a14_a13 and b24)))) xor ((((a26_a25_a24_a16_a15_a14 and b23) xor (a27_a26_a25_a17_a16_a15 and b22)) xor ((a28_a27_a26_a18_a17_a16 and b21) xor (a29_a28_a27_a19_a18_a17 and b20))) xor (((a30_a29_a28_a20_a19_a18 and b19) xor (a31_a30_a29_a21_a20_a19 and b18)) xor ((a31_a30_a22_a21_a20 and b17) xor (a31_a23_a22_a21 and b16))))) xor (((((a24_a23_a22 and b15) xor (a25_a24_a23 and b14)) xor ((a26_a25_a24 and b13) xor (a27_a26_a25 and b12))) xor (((a28_a27_a26 and b11) xor (a29_a28_a27 and b10)) xor ((a30_a29_a28 and b9) xor (a31_a30_a29 and b8)))) xor ((((a31_a30_a0 and b7) xor (a31_a1 and b6)) xor ((a2 and b5) xor (a3 and b4))) xor (((a4 and b3) xor (a5 and b2)) xor ((a6 and b1) xor (a7 and b0))))));
    ab6 <= ((((((a27_a26_a25_a17_a16_a15_a7_a6_a5 and b31) xor (a28_a27_a26_a18_a17_a16_a8_a7_a6 and b30)) xor ((a29_a28_a27_a19_a18_a17_a9_a8_a7 and b29) xor (a30_a29_a28_a20_a19_a18_a10_a9_a8 and b28))) xor (((a31_a30_a29_a21_a20_a19_a11_a10_a9 and b27) xor (a31_a30_a22_a21_a20_a12_a11_a10 and b26)) xor ((a31_a23_a22_a21_a13_a12_a11 and b25) xor (a24_a23_a22_a14_a13_a12 and b24)))) xor ((((a25_a24_a23_a15_a14_a13 and b23) xor (a26_a25_a24_a16_a15_a14 and b22)) xor ((a27_a26_a25_a17_a16_a15 and b21) xor (a28_a27_a26_a18_a17_a16 and b20))) xor (((a29_a28_a27_a19_a18_a17 and b19) xor (a30_a29_a28_a20_a19_a18 and b18)) xor ((a31_a30_a29_a21_a20_a19 and b17) xor (a31_a30_a22_a21_a20 and b16))))) xor (((((a31_a23_a22_a21 and b15) xor (a24_a23_a22 and b14)) xor ((a25_a24_a23 and b13) xor (a26_a25_a24 and b12))) xor (((a27_a26_a25 and b11) xor (a28_a27_a26 and b10)) xor ((a29_a28_a27 and b9) xor (a30_a29_a28 and b8)))) xor ((((a31_a30_a29 and b7) xor (a31_a30_a0 and b6)) xor ((a31_a1 and b5) xor (a2 and b4))) xor (((a3 and b3) xor (a4 and b2)) xor ((a5 and b1) xor (a6 and b0))))));
    ab5 <= ((((((a26_a25_a24_a16_a15_a14_a6_a5_a4 and b31) xor (a27_a26_a25_a17_a16_a15_a7_a6_a5 and b30)) xor ((a28_a27_a26_a18_a17_a16_a8_a7_a6 and b29) xor (a29_a28_a27_a19_a18_a17_a9_a8_a7 and b28))) xor (((a30_a29_a28_a20_a19_a18_a10_a9_a8 and b27) xor (a31_a30_a29_a21_a20_a19_a11_a10_a9 and b26)) xor ((a31_a30_a22_a21_a20_a12_a11_a10 and b25) xor (a31_a23_a22_a21_a13_a12_a11 and b24)))) xor ((((a24_a23_a22_a14_a13_a12 and b23) xor (a25_a24_a23_a15_a14_a13 and b22)) xor ((a26_a25_a24_a16_a15_a14 and b21) xor (a27_a26_a25_a17_a16_a15 and b20))) xor (((a28_a27_a26_a18_a17_a16 and b19) xor (a29_a28_a27_a19_a18_a17 and b18)) xor ((a30_a29_a28_a20_a19_a18 and b17) xor (a31_a30_a29_a21_a20_a19 and b16))))) xor (((((a31_a30_a22_a21_a20 and b15) xor (a31_a23_a22_a21 and b14)) xor ((a24_a23_a22 and b13) xor (a25_a24_a23 and b12))) xor (((a26_a25_a24 and b11) xor (a27_a26_a25 and b10)) xor ((a28_a27_a26 and b9) xor (a29_a28_a27 and b8)))) xor ((((a30_a29_a28 and b7) xor (a31_a30_a29 and b6)) xor ((a31_a30_a0 and b5) xor (a31_a1 and b4))) xor (((a2 and b3) xor (a3 and b2)) xor ((a4 and b1) xor (a5 and b0))))));
    ab4 <= ((((((a25_a24_a23_a15_a14_a13_a5_a4_a3 and b31) xor (a26_a25_a24_a16_a15_a14_a6_a5_a4 and b30)) xor ((a27_a26_a25_a17_a16_a15_a7_a6_a5 and b29) xor (a28_a27_a26_a18_a17_a16_a8_a7_a6 and b28))) xor (((a29_a28_a27_a19_a18_a17_a9_a8_a7 and b27) xor (a30_a29_a28_a20_a19_a18_a10_a9_a8 and b26)) xor ((a31_a30_a29_a21_a20_a19_a11_a10_a9 and b25) xor (a31_a30_a22_a21_a20_a12_a11_a10 and b24)))) xor ((((a31_a23_a22_a21_a13_a12_a11 and b23) xor (a24_a23_a22_a14_a13_a12 and b22)) xor ((a25_a24_a23_a15_a14_a13 and b21) xor (a26_a25_a24_a16_a15_a14 and b20))) xor (((a27_a26_a25_a17_a16_a15 and b19) xor (a28_a27_a26_a18_a17_a16 and b18)) xor ((a29_a28_a27_a19_a18_a17 and b17) xor (a30_a29_a28_a20_a19_a18 and b16))))) xor (((((a31_a30_a29_a21_a20_a19 and b15) xor (a31_a30_a22_a21_a20 and b14)) xor ((a31_a23_a22_a21 and b13) xor (a24_a23_a22 and b12))) xor (((a25_a24_a23 and b11) xor (a26_a25_a24 and b10)) xor ((a27_a26_a25 and b9) xor (a28_a27_a26 and b8)))) xor ((((a29_a28_a27 and b7) xor (a30_a29_a28 and b6)) xor ((a31_a30_a29 and b5) xor (a31_a30_a0 and b4))) xor (((a31_a1 and b3) xor (a2 and b2)) xor ((a3 and b1) xor (a4 and b0))))));
    ab3 <= ((((((a24_a23_a22_a14_a13_a12_a4_a3_a2 and b31) xor (a25_a24_a23_a15_a14_a13_a5_a4_a3 and b30)) xor ((a26_a25_a24_a16_a15_a14_a6_a5_a4 and b29) xor (a27_a26_a25_a17_a16_a15_a7_a6_a5 and b28))) xor (((a28_a27_a26_a18_a17_a16_a8_a7_a6 and b27) xor (a29_a28_a27_a19_a18_a17_a9_a8_a7 and b26)) xor ((a30_a29_a28_a20_a19_a18_a10_a9_a8 and b25) xor (a31_a30_a29_a21_a20_a19_a11_a10_a9 and b24)))) xor ((((a31_a30_a22_a21_a20_a12_a11_a10 and b23) xor (a31_a23_a22_a21_a13_a12_a11 and b22)) xor ((a24_a23_a22_a14_a13_a12 and b21) xor (a25_a24_a23_a15_a14_a13 and b20))) xor (((a26_a25_a24_a16_a15_a14 and b19) xor (a27_a26_a25_a17_a16_a15 and b18)) xor ((a28_a27_a26_a18_a17_a16 and b17) xor (a29_a28_a27_a19_a18_a17 and b16))))) xor (((((a30_a29_a28_a20_a19_a18 and b15) xor (a31_a30_a29_a21_a20_a19 and b14)) xor ((a31_a30_a22_a21_a20 and b13) xor (a31_a23_a22_a21 and b12))) xor (((a24_a23_a22 and b11) xor (a25_a24_a23 and b10)) xor ((a26_a25_a24 and b9) xor (a27_a26_a25 and b8)))) xor ((((a28_a27_a26 and b7) xor (a29_a28_a27 and b6)) xor ((a30_a29_a28 and b5) xor (a31_a30_a29 and b4))) xor (((a31_a30_a0 and b3) xor (a31_a1 and b2)) xor ((a2 and b1) xor (a3 and b0))))));
    ab2 <= ((((((a23_a22_a21_a13_a12_a11_a3_a2_a1 and b31) xor (a24_a23_a22_a14_a13_a12_a4_a3_a2 and b30)) xor ((a25_a24_a23_a15_a14_a13_a5_a4_a3 and b29) xor (a26_a25_a24_a16_a15_a14_a6_a5_a4 and b28))) xor (((a27_a26_a25_a17_a16_a15_a7_a6_a5 and b27) xor (a28_a27_a26_a18_a17_a16_a8_a7_a6 and b26)) xor ((a29_a28_a27_a19_a18_a17_a9_a8_a7 and b25) xor (a30_a29_a28_a20_a19_a18_a10_a9_a8 and b24)))) xor ((((a31_a30_a29_a21_a20_a19_a11_a10_a9 and b23) xor (a31_a30_a22_a21_a20_a12_a11_a10 and b22)) xor ((a31_a23_a22_a21_a13_a12_a11 and b21) xor (a24_a23_a22_a14_a13_a12 and b20))) xor (((a25_a24_a23_a15_a14_a13 and b19) xor (a26_a25_a24_a16_a15_a14 and b18)) xor ((a27_a26_a25_a17_a16_a15 and b17) xor (a28_a27_a26_a18_a17_a16 and b16))))) xor (((((a29_a28_a27_a19_a18_a17 and b15) xor (a30_a29_a28_a20_a19_a18 and b14)) xor ((a31_a30_a29_a21_a20_a19 and b13) xor (a31_a30_a22_a21_a20 and b12))) xor (((a31_a23_a22_a21 and b11) xor (a24_a23_a22 and b10)) xor ((a25_a24_a23 and b9) xor (a26_a25_a24 and b8)))) xor ((((a27_a26_a25 and b7) xor (a28_a27_a26 and b6)) xor ((a29_a28_a27 and b5) xor (a30_a29_a28 and b4))) xor (((a31_a30_a29 and b3) xor (a31_a30_a0 and b2)) xor ((a31_a1 and b1) xor (a2 and b0))))));
    ab1 <= ((((((a22_a21_a12_a11_a2_a1 and b31) xor (a23_a22_a13_a12_a3_a2 and b30)) xor ((a24_a23_a14_a13_a4_a3 and b29) xor (a25_a24_a15_a14_a5_a4 and b28))) xor (((a26_a25_a16_a15_a6_a5 and b27) xor (a27_a26_a17_a16_a7_a6 and b26)) xor ((a28_a27_a18_a17_a8_a7 and b25) xor (a29_a28_a19_a18_a9_a8 and b24)))) xor ((((a30_a29_a20_a19_a10_a9 and b23) xor (a31_a30_a21_a20_a11_a10 and b22)) xor ((a31_a22_a21_a12_a11 and b21) xor (a23_a22_a13_a12 and b20))) xor (((a24_a23_a14_a13 and b19) xor (a25_a24_a15_a14 and b18)) xor ((a26_a25_a16_a15 and b17) xor (a27_a26_a17_a16 and b16))))) xor (((((a28_a27_a18_a17 and b15) xor (a29_a28_a19_a18 and b14)) xor ((a30_a29_a20_a19 and b13) xor (a31_a30_a21_a20 and b12))) xor (((a31_a22_a21 and b11) xor (a23_a22 and b10)) xor ((a24_a23 and b9) xor (a25_a24 and b8)))) xor ((((a26_a25 and b7) xor (a27_a26 and b6)) xor ((a28_a27 and b5) xor (a29_a28 and b4))) xor (((a30_a29 and b3) xor (a31_a30 and b2)) xor ((a31_a0 and b1) xor (a1 and b0))))));
    ab0 <= ((((((a21_a11_a1 and b31) xor (a22_a12_a2 and b30)) xor ((a23_a13_a3 and b29) xor (a24_a14_a4 and b28))) xor (((a25_a15_a5 and b27) xor (a26_a16_a6 and b26)) xor ((a27_a17_a7 and b25) xor (a28_a18_a8 and b24)))) xor ((((a29_a19_a9 and b23) xor (a30_a20_a10 and b22)) xor ((a31_a21_a11 and b21) xor (a22_a12 and b20))) xor (((a23_a13 and b19) xor (a24_a14 and b18)) xor ((a25_a15 and b17) xor (a26_a16 and b16))))) xor (((((a27_a17 and b15) xor (a28_a18 and b14)) xor ((a29_a19 and b13) xor (a30_a20 and b12))) xor (((a31_a21 and b11) xor (a22 and b10)) xor ((a23 and b9) xor (a24 and b8)))) xor ((((a25 and b7) xor (a26 and b6)) xor ((a27 and b5) xor (a28 and b4))) xor (((a29 and b3) xor (a30 and b2)) xor ((a31 and b1) xor (a0 and b0))))));
end architecture mxl_mastrovito;