-------------------------------------------------------------------------------
--! @file       TriviA0v2_SC.vhd
--! @brief      Stream cipher portion of TriviA0v2
--! @project    CAESAR Candidate Evaluation
--! @author     Michael X. Lyons
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
--! Bit numbering follows the convention in the TriviA0v2 specification:
--!     bits are numbered from 1 up to the length.
--! The stream state consists of 3 non-linear feedback shift registers A, B, C.
--!     Their lengths are 132, 105, and 147 bits respectively (totalling 384 = 128 x 3).
-------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;


entity TriviA0v2_SC is
    port (
         clk      : in  std_logic;                   -- clock
         en       : in  std_logic;                   -- enable
         load     : in  std_logic;                   -- initialize the stream state and load 'key' and 'iv' into it
         key      : in  std_logic_vector(1 to 128);  -- key to be loaded
         iv       : in  std_logic_vector(1 to 128);  -- IV  to be loaded
         insert   : in  std_logic;                   -- insert a tag into the stream state
         tag      : in  std_logic_vector(1 to 128);  -- tag to be inserted
         keyExt64 : out std_logic_vector(1 to  64);  -- also known as K^v "variable key"
         stExt64  : out std_logic_vector(1 to  64)   -- also known as K^t "tag key"
		 );
end entity TriviA0v2_SC;

architecture behavioral of TriviA0v2_SC is
    -- the stream state NLFSRs
    signal a_reg, a_next : std_logic_vector(1 to 132);  -- LFSR A
    signal b_reg, b_next : std_logic_vector(1 to 105);  -- LFSR B
    signal c_reg, c_next : std_logic_vector(1 to 147);  -- LFSR C

    -- 64-bit shift values
    signal t1, t2, t3    :  std_logic_vector(1 to 64);

    -- intermediate values used in shift logic, output logic
    signal t1a, t1b, t1c :  std_logic_vector(1 to 64);
begin
    -- registers
    process (clk)
    begin
        if rising_edge(clk) then
            if en = '1' then
                if load = '1' then
                    -- load key and IV into initialized stream state

                    -- this is always followed by 18 cycles of stream state update
                    -- we could combine this with the first of those 18 cycles
                    a_reg(  1 to 128) <= key;
                    a_reg(129 to 132) <= (others => '0');

                    b_reg(  1 to 102) <= (others => '0');
                    b_reg(103 to 105) <= "111";

                    c_reg(  1 to 128) <= iv;
                    c_reg(129 to 147) <= (others => '0');
                elsif insert = '1' then
                    -- XOR tag into the first part of the stream state

                    -- this is also always followed by 18 cycles of stream state update
                    -- we could combine this with the first of those 18 cycles

                    -- XOR the tag into the first 128 bits of the A register
                    a_reg(  1 to 128) <= a_reg(  1 to 128) xor tag;
                    a_reg(129 to 132) <= a_reg(129 to 132);
                else
                    -- cycle the stream state
                    a_reg <= a_next;
                    b_reg <= b_next;
                    c_reg <= c_next;
                end if;
            end if;
        end if;
    end process;

    -- next-state logic
    -- cycle the stream cipher

    -- these values are used for both shifting in and extracting
    t1a <= a_reg(  3 to  66) xor
           a_reg( 69 to 132);

    t1b <= b_reg(  6 to  69) xor
           b_reg( 42 to 105);

    t1c <= c_reg(  3 to  66) xor
           c_reg( 84 to 147);

    -- data to shift in    
    -- t1 <= t1a xor
    --       (a_reg( 67 to 130) and a_reg( 68 to 131)) xor
    --       b_reg( 33 to  96);

    -- t2 <= t1b xor
    --       (b_reg( 40 to 103) and b_reg( 41 to 104)) xor
    --       c_reg( 57 to 120);

    -- t3 <= t1c xor
    --       (c_reg( 82 to 145) and c_reg( 83 to 146)) xor
    --       a_reg( 12 to  75);

    t1 <= (t1a xor b_reg( 33 to  96))
          xor
          (a_reg( 67 to 130) and a_reg( 68 to 131));

    t2 <= (t1b xor c_reg( 57 to 120))
          xor
          (b_reg( 40 to 103) and b_reg( 41 to 104));

    t3 <= (t1c xor a_reg( 12 to  75))
          xor
          (c_reg( 82 to 145) and c_reg( 83 to 146));

    -- 64 bits to be shifted into each register in the next clock cycle
    -- a_next <= t3 & a_reg(  1 to  68);
    -- b_next <= t1 & b_reg(  1 to  41);
    -- c_next <= t2 & c_reg(  1 to  83);
    a_next(  1 to  64) <= t3;
    a_next( 65 to 132) <= a_reg(  1 to  68);

    b_next(  1 to  64) <= t1;
    b_next( 65 to 105) <= b_reg(  1 to  41);

    c_next(  1 to  64) <= t2;
    c_next( 65 to 147) <= c_reg(  1 to  83);


    -- outputs
    -- keyExt64 <= t1a xor
    --             t1b xor
    --             t1c xor
    --             (a_reg( 39 to 102) and b_reg(  3 to  66));
    keyExt64 <= (t1a xor t1b)
                xor
                (t1c xor (a_reg( 39 to 102) and b_reg(  3 to  66)));

    stExt64  <= a_reg(1 to 64);
end architecture behavioral;